@@ -1,11 +1,12 @@
-# $Id: LibXML.pm,v 1.99 2004/02/14 18:35:14 pajas Exp $
+# $Id: LibXML.pm,v 1.109 2006/07/27 12:08:20 pajas Exp $
 
 package XML::LibXML;
 
 use strict;
 use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS
             $skipDTD $skipXMLDeclaration $setTagCompression
-            $MatchCB $ReadCB $OpenCB $CloseCB );
+            $MatchCB $ReadCB $OpenCB $CloseCB 
+            );
 use Carp;
 
 use XML::LibXML::Common qw(:encoding :libxml);
@@ -14,7 +15,7 @@ use XML::LibXML::NodeList;
 use IO::Handle; # for FH reads called as methods
 
 
-$VERSION = "1.57";
+$VERSION = "1.59";
 require Exporter;
 require DynaLoader;
 
@@ -121,6 +122,8 @@ sub new {
         $self->set_handler( $options{Handler} );
     }
 
+    $self->{XML_LIBXML_EXT_DTD} = 1;
+
     return $self;
 }
 
@@ -147,10 +150,24 @@ sub createDocument {
 #-------------------------------------------------------------------------#
 # callback functions                                                      #
 #-------------------------------------------------------------------------#
+
+sub input_callbacks {
+    my $self     = shift;
+    my $icbclass = shift;
+
+    if ( defined $icbclass ) {
+        $self->{XML_LIBXML_CALLBACK_STACK} = $icbclass;
+    }
+    return $self->{XML_LIBXML_CALLBACK_STACK};
+}
+
 sub match_callback {
     my $self = shift;
     if ( ref $self ) {
-        $self->{XML_LIBXML_MATCH_CB} = shift if scalar @_;
+        if ( scalar @_ ) {
+            $self->{XML_LIBXML_MATCH_CB} = shift;
+            $self->{XML_LIBXML_CALLBACK_STACK} = undef;
+        }
         return $self->{XML_LIBXML_MATCH_CB};
     }
     else {
@@ -162,7 +179,10 @@ sub match_callback {
 sub read_callback {
     my $self = shift;
     if ( ref $self ) {
-        $self->{XML_LIBXML_READ_CB} = shift if scalar @_;
+        if ( scalar @_ ) {
+            $self->{XML_LIBXML_READ_CB} = shift;
+            $self->{XML_LIBXML_CALLBACK_STACK} = undef;
+        }
         return $self->{XML_LIBXML_READ_CB};
     }
     else {
@@ -174,7 +194,10 @@ sub read_callback {
 sub close_callback {
     my $self = shift;
     if ( ref $self ) {
-        $self->{XML_LIBXML_CLOSE_CB} = shift if scalar @_;
+        if ( scalar @_ ) {
+            $self->{XML_LIBXML_CLOSE_CB} = shift;
+            $self->{XML_LIBXML_CALLBACK_STACK} = undef;
+        }
         return $self->{XML_LIBXML_CLOSE_CB};
     }
     else {
@@ -186,7 +209,10 @@ sub close_callback {
 sub open_callback {
     my $self = shift;
     if ( ref $self ) {
-        $self->{XML_LIBXML_OPEN_CB} = shift if scalar @_;
+        if ( scalar @_ ) {
+            $self->{XML_LIBXML_OPEN_CB} = shift;
+            $self->{XML_LIBXML_CALLBACK_STACK} = undef;
+        }
         return $self->{XML_LIBXML_OPEN_CB};
     }
     else {
@@ -201,6 +227,7 @@ sub callbacks {
         if (@_) {
             my ($match, $open, $read, $close) = @_;
             @{$self}{qw(XML_LIBXML_MATCH_CB XML_LIBXML_OPEN_CB XML_LIBXML_READ_CB XML_LIBXML_CLOSE_CB)} = ($match, $open, $read, $close);
+            $self->{XML_LIBXML_CALLBACK_STACK} = undef;
         }
         else {
             return @{$self}{qw(XML_LIBXML_MATCH_CB XML_LIBXML_OPEN_CB XML_LIBXML_READ_CB XML_LIBXML_CLOSE_CB)};
@@ -231,6 +258,13 @@ sub recover {
     return $self->{XML_LIBXML_RECOVER};
 }
 
+sub recover_silently {
+    my $self = shift;
+    my $arg = shift;
+    (($arg == 1) ? $self->recover(2) : $self->recover($arg)) if defined($arg);
+    return ($self->recover() == 2) ? 1 : 0;
+}
+
 sub expand_entities {
     my $self = shift;
     $self->{XML_LIBXML_EXPAND_ENTITIES} = shift if scalar @_;
@@ -285,6 +319,11 @@ sub gdome_dom {
     return $self->{XML_LIBXML_GDOME};
 }
 
+sub clean_namespaces {
+    my $self = shift;
+    $self->{XML_LIBXML_NSCLEAN} = shift if scalar @_;
+    return $self->{XML_LIBXML_NSCLEAN};
+}
 
 #-------------------------------------------------------------------------#
 # set the optional SAX(2) handler                                         #
@@ -320,6 +359,7 @@ sub _auto_expand {
         my $err = $@;
         $self->{_State_} = 0;
         if ($err) {
+            $self->_cleanup_callbacks();
             $result = undef;
             croak $err;
         }
@@ -327,6 +367,34 @@ sub _auto_expand {
     return $result;
 }
 
+sub _init_callbacks {
+    my $self = shift;
+    my $icb = $self->{XML_LIBXML_CALLBACK_STACK};
+    
+    unless ( defined $icb ) {
+        $self->{XML_LIBXML_CALLBACK_STACK} = XML::LibXML::InputCallback->new();
+        $icb = $self->{XML_LIBXML_CALLBACK_STACK};
+    }
+
+    my $mcb = $self->match_callback();
+    my $ocb = $self->open_callback();
+    my $rcb = $self->read_callback();
+    my $ccb = $self->close_callback();
+
+    if ( defined $mcb and defined $ocb and defined $rcb and defined $ccb ) {
+        $icb->register_callbacks( [$mcb, $ocb, $rcb, $ccb] );
+    }
+    
+    $icb->init_callbacks();
+}
+
+sub _cleanup_callbacks {
+    my $self = shift;
+    $self->{XML_LIBXML_CALLBACK_STACK}->cleanup_callbacks();
+    my $mcb = $self->match_callback();
+    $self->{XML_LIBXML_CALLBACK_STACK}->unregister_callbacks( [$mcb] );
+}
+
 sub __read {
     read($_[0], $_[1], $_[2]);
 }
@@ -349,6 +417,7 @@ sub __write {
 #-------------------------------------------------------------------------#
 sub parse_string {
     my $self = shift;
+    croak("parse_string is not a class method! Create a parser object with XML::LibXML->new first!") unless ref $self;
     croak("parse already in progress") if $self->{_State_};
 
     unless ( defined $_[0] and length $_[0] ) {
@@ -358,14 +427,18 @@ sub parse_string {
     $self->{_State_} = 1;
     my $result;
 
+    $self->_init_callbacks();
+
     if ( defined $self->{SAX} ) {
         my $string = shift;
         $self->{SAX_ELSTACK} = [];
+        
         eval { $result = $self->_parse_sax_string($string); };
 
         my $err = $@;
         $self->{_State_} = 0;
         if ($err) {
+            $self->_cleanup_callbacks();
             croak $err;
         }
     }
@@ -375,26 +448,33 @@ sub parse_string {
         my $err = $@;
         $self->{_State_} = 0;
         if ($err) {
+            $self->_cleanup_callbacks();
             croak $err;
         }
 
         $result = $self->_auto_expand( $result, $self->{XML_LIBXML_BASE_URI} );
     }
+    $self->_cleanup_callbacks();
 
     return $result;
 }
 
 sub parse_fh {
     my $self = shift;
+    croak("parse_fh is not a class method! Create a parser object with XML::LibXML->new first!") unless ref $self;
     croak("parse already in progress") if $self->{_State_};
     $self->{_State_} = 1;
     my $result;
+
+    $self->_init_callbacks();
+
     if ( defined $self->{SAX} ) {
         $self->{SAX_ELSTACK} = [];
         eval { $self->_parse_sax_fh( @_ );  };
         my $err = $@;
         $self->{_State_} = 0;
         if ($err) {
+            $self->_cleanup_callbacks();
             croak $err;
         }
     }
@@ -403,26 +483,34 @@ sub parse_fh {
         my $err = $@;
         $self->{_State_} = 0;
         if ($err) {
+            $self->_cleanup_callbacks();
             croak $err;
         }
 
         $result = $self->_auto_expand( $result, $self->{XML_LIBXML_BASE_URI} );
     }
 
+    $self->_cleanup_callbacks();
+
     return $result;
 }
 
 sub parse_file {
     my $self = shift;
+    croak("parse_file is not a class method! Create a parser object with XML::LibXML->new first!") unless ref $self;
     croak("parse already in progress") if $self->{_State_};
     $self->{_State_} = 1;
     my $result;
+
+    $self->_init_callbacks();
+
     if ( defined $self->{SAX} ) {
         $self->{SAX_ELSTACK} = [];
         eval { $self->_parse_sax_file( @_ );  };
         my $err = $@;
         $self->{_State_} = 0;
         if ($err) {
+            $self->_cleanup_callbacks();
             croak $err;
         }
     }
@@ -431,11 +519,13 @@ sub parse_file {
         my $err = $@;
         $self->{_State_} = 0;
         if ($err) {
+            $self->_cleanup_callbacks();
             croak $err;
         }
 
         $result = $self->_auto_expand( $result );
     }
+    $self->_cleanup_callbacks();
 
     return $result;
 }
@@ -445,6 +535,7 @@ sub parse_xml_chunk {
     # max 2 parameter:
     # 1: the chunk
     # 2: the encoding of the string
+    croak("parse_xml_chunk is not a class method! Create a parser object with XML::LibXML->new first!") unless ref $self;
     croak("parse already in progress") if $self->{_State_};    my $result;
 
     unless ( defined $_[0] and length $_[0] ) {
@@ -452,6 +543,9 @@ sub parse_xml_chunk {
     }
 
     $self->{_State_} = 1;
+
+    $self->_init_callbacks();
+
     if ( defined $self->{SAX} ) {
         eval {
             $self->_parse_sax_xml_chunk( @_ );
@@ -469,6 +563,8 @@ sub parse_xml_chunk {
         eval { $result = $self->_parse_xml_chunk( @_ ); };
     }
 
+    $self->_cleanup_callbacks();
+
     my $err = $@;
     $self->{_State_} = 0;
     if ($err) {
@@ -480,23 +576,130 @@ sub parse_xml_chunk {
 
 sub parse_balanced_chunk {
     my $self = shift;
-    return $self->parse_xml_chunk( @_ );
+    $self->_init_callbacks();
+    my $rv;
+    eval {
+        $rv = $self->parse_xml_chunk( @_ );
+    };
+    $self->_cleanup_callbacks();
+    if ( $@ ) {
+        croak $@;
+    }
+    return $rv
 }
 
 # java style
 sub processXIncludes {
     my $self = shift;
     my $doc = shift;
-    return $self->_processXIncludes($doc || " ");
+    if ( $self->{_State_} != 1 ) {
+        $self->_init_callbacks();
+    }
+    my $rv;
+    eval {
+        $rv = $self->_processXIncludes($doc || " ");
+    };
+
+    if ( $self->{_State_} != 1 ) {
+        $self->_cleanup_callbacks();
+    }
+
+    if ( $@ ) {
+        croak $@;
+    }
+    return $rv;
 }
 
 # perl style
 sub process_xincludes {
     my $self = shift;
     my $doc = shift;
-    return $self->_processXIncludes($doc || " ");
+    my $rv;
+    $self->_init_callbacks();
+    eval {
+        $rv = $self->_processXIncludes($doc || " ");
+    };
+    $self->_cleanup_callbacks();
+    if ( $@ ) {
+        croak $@;
+    }
+    return $rv;
 }
 
+#-------------------------------------------------------------------------#
+# HTML parsing functions                                                  #
+#-------------------------------------------------------------------------#
+
+sub parse_html_string {
+    my $self = shift;
+    croak("parse_html_string is not a class method! Create a parser object with XML::LibXML->new first!") unless ref $self;
+    croak("parse already in progress") if $self->{_State_};
+
+    unless ( defined $_[0] and length $_[0] ) {
+        croak("Empty String");
+    }
+
+    $self->{_State_} = 1;
+    my $result;
+
+    $self->_init_callbacks();
+
+    eval { $result = $self->_parse_html_string( @_ ); };
+    
+    my $err = $@;
+    $self->{_State_} = 0;
+    if ($err) {
+      $self->_cleanup_callbacks();
+      croak $err;
+    }
+        
+    $self->_cleanup_callbacks();
+
+    return $result;
+}
+
+sub parse_html_file {
+    my $self = shift;
+    croak("parse_html_file is not a class method! Create a parser object with XML::LibXML->new first!") unless ref $self;
+    croak("parse already in progress") if $self->{_State_};
+    $self->{_State_} = 1;
+    my $result;
+
+    $self->_init_callbacks();
+    
+    eval { $result = $self->_parse_html_file(@_); };
+    my $err = $@;
+    $self->{_State_} = 0;
+    if ($err) {
+      $self->_cleanup_callbacks();
+      croak $err;
+    }
+    
+    $self->_cleanup_callbacks();
+
+    return $result;
+}
+
+sub parse_html_fh {
+    my $self = shift;
+    croak("parse_html_fh is not a class method! Create a parser object with XML::LibXML->new first!") unless ref $self;
+    croak("parse already in progress") if $self->{_State_};
+    $self->{_State_} = 1;
+    my $result;
+
+    $self->_init_callbacks();
+    
+    eval { $result = $self->_parse_html_fh( @_ ); };
+    my $err = $@;
+    $self->{_State_} = 0;
+    if ($err) {
+      $self->_cleanup_callbacks();
+      croak $err;
+    }
+    $self->_cleanup_callbacks();
+
+    return $result;
+}
 
 #-------------------------------------------------------------------------#
 # push parser interface                                                   #
@@ -519,12 +722,21 @@ sub init_push {
 sub push {
     my $self = shift;
 
+    $self->_init_callbacks();
+    
     if ( not defined $self->{CONTEXT} ) {
         $self->init_push();
     }
 
-    foreach ( @_ ) {
-        $self->_push( $self->{CONTEXT}, $_ );
+    eval {
+        foreach ( @_ ) {
+            $self->_push( $self->{CONTEXT}, $_ );
+        }
+    };
+
+    $self->_cleanup_callbacks();
+    if ( $@ ) {
+        croak $@;
     }
 }
 
@@ -858,8 +1070,7 @@ sub getChildrenByTagName {
 
 sub getChildrenByTagNameNS {
     my ( $node, $nsURI, $name ) = @_;
-    my $xpath = "*[local-name()='$name' and namespace-uri()='$nsURI']";
-    my @nodes = $node->_findnodes($xpath);
+    my @nodes = $node->_getChildrenByTagNameNS($nsURI,$name);
     return wantarray ? @nodes : XML::LibXML::NodeList->new(@nodes);
 }
 
@@ -942,6 +1153,10 @@ package XML::LibXML::CDATASection;
 use vars qw(@ISA);
 @ISA     = ('XML::LibXML::Text');
 
+sub nodeName {
+    return "cdata";
+}
+
 1;
 
 #-------------------------------------------------------------------------#
@@ -1210,4 +1425,163 @@ sub new {
 
 1;
 
+#-------------------------------------------------------------------------#
+# XML::LibXML::InputCallback Interface                                    #
+#-------------------------------------------------------------------------#
+package XML::LibXML::InputCallback;
+
+use vars qw($_CUR_CB @_GLOBAL_CALLBACKS @_CB_STACK);
+
+$_CUR_CB = undef;
+
+@_GLOBAL_CALLBACKS = ();
+@_CB_STACK = ();
+
+#-------------------------------------------------------------------------#
+# global callbacks                                                        #
+#-------------------------------------------------------------------------#
+sub _callback_match {
+    my $uri = shift;
+    my $retval = 0;
+
+    # loop through the callbacks and and find the first matching
+    # The callbacks are stored in execution order (reverse stack order)
+    # any new global callbacks are shifted to the callback stack.
+    foreach my $cb ( @_GLOBAL_CALLBACKS ) {
+
+        # callbacks have to return 1, 0 or undef, while 0 and undef 
+        # are handled the same way. 
+        # in fact, if callbacks return other values, the global match 
+        # assumes silently that the callback failed.
+
+        $retval = $cb->[0]->($uri);
+
+        if ( defined $retval and $retval == 1 ) {
+            # make the other callbacks use this callback
+            $_CUR_CB = $cb;
+            unshift @_CB_STACK, $cb; 
+            last;
+        }
+    }
+
+    return $retval;
+}
+
+sub _callback_open {
+    my $uri = shift;
+    my $retval = undef;
+    
+    # the open callback has to return a defined value. 
+    # if one works on files this can be a file handle. But 
+    # depending on the needs of the callback it also can be a 
+    # database handle or a integer labeling a certain dataset.
+
+    if ( defined $_CUR_CB ) {
+        $retval = $_CUR_CB->[1]->( $uri );
+        
+        # reset the callbacks, if one callback cannot open an uri
+        if ( not defined $retval or $retval == 0 ) {
+            shift @_CB_STACK;
+            $_CUR_CB = $_CB_STACK[0];
+        }
+    }
+    
+    return $retval;
+}
+
+sub _callback_read {
+    my $fh = shift;
+    my $buflen = shift;
+
+    my $retval = undef;
+
+    if ( defined $_CUR_CB ) {
+        $retval = $_CUR_CB->[2]->( $fh, $buflen );
+    }
+
+    return $retval;
+}
+
+sub _callback_close {
+    my $fh = shift;
+    my $retval = 0;
+    
+    if ( defined $_CUR_CB ) {
+        $retval = $_CUR_CB->[3]->( $fh );
+        shift @_CB_STACK;
+        $_CUR_CB = $_CB_STACK[0];
+    }
+
+    return $retval;
+}
+
+#-------------------------------------------------------------------------#
+# member functions and methods                                            #
+#-------------------------------------------------------------------------#
+
+sub new {
+    my $CLASS = shift;
+    return bless {'_CALLBACKS' => []}, $CLASS;
+}
+
+# add a callback set to the callback stack
+# synopsis: $icb->register_callbacks( [$match_cb, $open_cb, $read_cb, $close_cb] );
+sub register_callbacks {
+    my $self = shift;
+    my $cbset = shift;
+    
+    # test if callback set is complete
+    if ( ref $cbset eq "ARRAY" and scalar( @$cbset ) == 4 ) {
+        unshift @{$self->{_CALLBACKS}}, $cbset;
+    }
+}
+
+# remove a callback set to the callback stack
+# if a callback set is passed, this function will check for the match function
+sub unregister_callbacks {
+    my $self = shift;
+    my $cbset = shift;
+    if ( ref $cbset eq "ARRAY" and scalar( @$cbset ) == 4 ) {
+        $self->{_CALLBACKS} = [grep { $_->[0] != $cbset->[0] } @{$self->{_CALLBACKS}}];
+    }
+    else {
+        shift @{$self->{_CALLBACKS}};
+    }
+}
+
+# make libxml2 use the callbacks 
+sub init_callbacks {
+    my $self = shift;
+
+    $_CUR_CB           = undef;
+    @_CB_STACK         = ();
+
+    @_GLOBAL_CALLBACKS = @{ $self->{_CALLBACKS} };
+    
+    if ( defined $XML::LibXML::match_cb and 
+         defined $XML::LibXML::open_cb  and 
+         defined $XML::LibXML::read_cb  and 
+         defined $XML::LibXML::close_cb ) {
+        push @_GLOBAL_CALLBACKS, [$XML::LibXML::match_cb,
+                                  $XML::LibXML::open_cb,
+                                  $XML::LibXML::read_cb,
+                                  $XML::LibXML::close_cb];
+    }
+
+    $self->lib_init_callbacks();
+}
+
+# reset libxml2's callbacks
+sub cleanup_callbacks {
+    my $self = shift;
+    
+    $_CUR_CB           = undef;
+    @_GLOBAL_CALLBACKS = ();
+    @_CB_STACK         = ();
+
+    $self->lib_cleanup_callbacks();
+}
+
+1;
+
 __END__
