@@ -1,5 +1,5 @@
-# $Id: SAX.pm,v 1.3 2002/06/21 11:33:35 phish Exp $
-
+# $Id: SAX.pm,v 1.9 2006/06/22 07:34:07 pajas Exp $
+# Copyright (c) 2001-2002, AxKit.com Ltd. All rights reserved.
 package XML::LibXML::SAX;
 
 use strict;
@@ -10,7 +10,7 @@ $VERSION = '1.00';
 use XML::LibXML;
 use XML::SAX::Base;
 
-@ISA = qw(XML::SAX::Base);
+use base qw(XML::SAX::Base);
 
 use Carp;
 use IO::File;
@@ -32,7 +32,8 @@ sub _parse_bytestream {
 
 sub _parse_string {
     my ( $self, $string ) = @_;
-    $self->{ParserOptions}{LibParser}      = XML::LibXML->new;
+#    $self->{ParserOptions}{LibParser}      = XML::LibXML->new;
+    $self->{ParserOptions}{LibParser}      = XML::LibXML->new()     unless defined $self->{ParserOptions}{LibParser};
     $self->{ParserOptions}{ParseFunc}      = \&XML::LibXML::parse_string;
     $self->{ParserOptions}{ParseFuncParam} = $string;
     return $self->_parse;
@@ -59,13 +60,22 @@ sub _parse {
     my $args = bless $self->{ParserOptions}, ref($self);
 
     $args->{LibParser}->set_handler( $self );
-    $args->{ParseFunc}->($args->{LibParser}, $args->{ParseFuncParam});
+    eval {
+      $args->{ParseFunc}->($args->{LibParser}, $args->{ParseFuncParam});
+    };
 
     if ( $args->{LibParser}->{SAX}->{State} == 1 ) {
         croak( "SAX Exception not implemented, yet; Data ended before document ended\n" );
     }
 
-    return $self->end_document({}); }
+    # break a possible circular reference    
+    $args->{LibParser}->set_handler( undef );
+    if ( $@ ) {
+        croak $@;
+    }
+    return $self->end_document({});
+}
+
 
 1;
 
