@@ -1,6 +1,6 @@
 /**
  * perl-libxml-sax.h
- * $Id: perl-libxml-sax.h,v 1.2 2002/05/20 10:36:49 phish Exp $
+ * $Id: perl-libxml-sax.h,v 1.4 2006/07/28 10:44:05 pajas Exp $
  */
 
 #ifndef __PERL_LIBXML_SAX_H__
@@ -18,10 +18,10 @@ extern "C" {
 
 /* has to be called in BOOT sequence */
 void
-PmmSAXInitialize();
+PmmSAXInitialize(pTHX);
 
 void
-PmmSAXInitContext( xmlParserCtxtPtr ctxt, SV * parser );
+PmmSAXInitContext( xmlParserCtxtPtr ctxt, SV * parser, SV * saved_error );
 
 void 
 PmmSAXCloseContext( xmlParserCtxtPtr ctxt );
