@@ -12,6 +12,7 @@ perlDocumentFunction(xmlXPathParserContextPtr ctxt, int nargs){
     xmlXPathObjectPtr obj = NULL, obj2 = NULL;
     xmlChar *base = NULL, *URI = NULL;
 
+
     if ((nargs < 1) || (nargs > 2)) {
         ctxt->error = XPATH_INVALID_ARITY;
         return;
@@ -30,6 +31,11 @@ perlDocumentFunction(xmlXPathParserContextPtr ctxt, int nargs){
         obj2 = valuePop(ctxt);
     }
 
+
+    /* first assure the XML::LibXML error handler is deactivated 
+       otherwise strange things might happen
+     */
+
     if (ctxt->value->type == XPATH_NODESET) {
         int i;
         xmlXPathObjectPtr newobj, ret;
@@ -60,6 +66,9 @@ perlDocumentFunction(xmlXPathParserContextPtr ctxt, int nargs){
         if (obj2 != NULL)
             xmlXPathFreeObject(obj2);
         valuePush(ctxt, ret);
+
+        /* reset the error old error handler before leaving
+         */
         return;
     }
     /*
@@ -70,6 +79,10 @@ perlDocumentFunction(xmlXPathParserContextPtr ctxt, int nargs){
         ctxt->error = XPATH_INVALID_TYPE;
         if (obj2 != NULL)
             xmlXPathFreeObject(obj2);
+
+        /* reset the error old error handler before leaving
+         */
+
         return;
     }
     obj = valuePop(ctxt);
@@ -114,6 +127,9 @@ perlDocumentFunction(xmlXPathParserContextPtr ctxt, int nargs){
     xmlXPathFreeObject(obj);
     if (obj2 != NULL)
         xmlXPathFreeObject(obj2);
+
+    /* reset the error old error handler before leaving
+     */
 }
 
 
@@ -194,11 +210,11 @@ domXPathFind( xmlNodePtr refNode, xmlChar * path ) {
         if ( tdoc != NULL ) {
             /* after looking through a fragment, we need to drop the
                fake document again */
-            froot->doc = NULL;
+            xmlSetTreeDoc(froot, NULL);
             tdoc->children = NULL;
             tdoc->last     = NULL;
-            froot->parent  = NULL;
-            refNode->doc = NULL;
+            /* next line is not required anymore */
+            /* refNode->doc = NULL; */
 
             xmlFreeDoc( tdoc );
         }
